; Copyright (C) 2023  Val Lorentz
; 
; This program is free software: you can redistribute it and/or modify it under the
; terms of the GNU Affero General Public License version 3, as published by the Free
; Software Foundation.
; 
; This program is distributed in the hope that it will be useful, but WITHOUT ANY
; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
; PARTICULAR PURPOSE.  See the GNU Affero General Public License for more details.
; 
; You should have received a copy of the GNU Affero General Public License along with
; this program.  If not, see <http://www.gnu.org/licenses/>.

(ns jable.cli
  (:require [jepsen [cli :as cli]
                    [tests :as tests]
                    [checker :as checker]
                    [generator :as gen]
                    [nemesis :as nemesis]]
            [jepsen.os.debian :as debian]
            [knossos.model :as model]
            [jable.client :as client]
            [jable.db :as db])
  (:import (knossos.model Model)))

(defn r   [_ _] {:type :invoke, :f :read-topic, :value nil})
(defn w   [_ _] {:type :invoke, :f :write-topic, :value (str (rand-int 5))})

(defrecord Channel [value]
  Model
  (step [c op]
    (case (:f op)
      :write-topic (Channel. (:value op))
      :read-topic  (if (= value (:value op))
                     c
                     (model/inconsistent
                       (str (pr-str value) "≠" (pr-str (:value op)))))))

  Object
  (toString [r] (pr-str value)))

(defn channel []
  (Channel. nil))

(defn sable-test
  "Given an options map from the command line runner (e.g. :nodes, :ssh,
  :concurrency, ...), constructs a test map."
  [opts]
  (merge tests/noop-test
         opts
         {:pure-generators true
          :name "etcd"
          :os   debian/os
          :db   (db/sable (:sable-bin opts) (:nodes opts))
          :client (client/jable.client.Client. nil nil nil)
          :nemesis         (nemesis/partition-random-halves)
          :checker         (checker/linearizable
                             {:model     (channel)
                              :algorithm :linear})
          :generator       (->> (gen/mix [r w])
                                (gen/stagger 1)
                                (gen/nemesis nil)
                                (gen/time-limit 15))}))

(def cli-opts
  "Additional command line options."
  [[nil "--sable-bin SABLE/BIN/" "Path to a directory containing Sable's binary (eg. ~/sable/target/debug/)"]])

(defn -main
  "Handles command line arguments. Can either run a test, or a web server for
  browsing results."
  [& args]
  (cli/run! (merge (cli/single-test-cmd {:test-fn sable-test
                                         :opt-spec cli-opts})
                   (cli/serve-cmd))
            args))
